#!/usr/bin/env python3

import parser_gau as pgau
import force_constant_mod as fc
import numpy as np
import average_across_types as aat
import readin_opts as rdin
import os

def print_GauHarm(*args):
    ele_ls, tp_ls, coord, bond_tp_ls, k_bond_ls, bond_ln_ls, angle_tp_ls, k_angle_ls, \
    angle_ln_ls, torsion_type_list, hybrid_list, chg, formal_chg, multi = args

    header_gjf = """%mem=1GB
%nprocshared=1
%chk=GauHarm.chk
#p Amber=(SoftOnly,Print) nosymm  Freq=intmodes
 
Title

{f_chg} {mult}
"""

    master_func = """
! Master function
NonBon 3 1 0 0 0.000 0.000 0.500 0.000 0.000 -1.2
"""
    fname = 'GauHarm.gjf'
    with open(fname, 'w') as fout:
        fout.write(header_gjf.format(f_chg = formal_chg, mult=multi))
        for m, p, l in zip(ele_ls, tp_ls, coord):
            s2 = '  '.join(f'{x:.6f}' for x in l)
            fout.write(f'{m}-{p}-0.00  {s2}\n')

        fout.write(master_func)
        fout.write('! SMARTFIELD FF\n')
        fout.write('!Bonds\n')
        for bond_tp, k_bond, bond_ln in zip(bond_tp_ls, k_bond_ls, bond_ln_ls):
            fout.write(f'HrmStr1 {bond_tp}  1.0 {bond_ln:.3f}\n')

        fout.write('!Angles\n')
        for angle_tp, k_angle, angle_ln in zip(angle_tp_ls, k_angle_ls, angle_ln_ls):
            fout.write(f'HrmBnd1 {angle_tp}  1.0 {angle_ln:.3f}\n')
            
        fout.write('!Torsions\n')
        for i, (torsion_type, hybrid) in enumerate(zip(torsion_type_list, hybrid_list)):
            if hybrid == 4.0 or hybrid == 2.0:
               formatted_phase = '0 180 0 0 '
               fout.write(f'AmbTrs {torsion_type} {formatted_phase} 0. {1.0} 0. 0. {float(hybrid)}\n')
            else:
               formatted_phase = '0 0 0 0 '
               fout.write(f'AmbTrs {torsion_type} {formatted_phase} 0. 0. {1.0} 0. {float(hybrid)}\n')
            
            
        fout.write('\n')
        
        
        
def print_GauNonBon(*args):
    elements, types, coordinates, bond_types, bond_lengths, angle_types, \
    angle_lengths, torsion_type_list, hybrid_list,  charges, vdw, formal_chg, multi = args
    header_gjf = """%mem=1GB
%nprocshared=1
%chk=GauNonBon.chk
#p Amber=(SoftOnly,Print) nosymm Freq=intmodes
 
Title

{f_chg} {mult}
"""

    master_func = """
! Master function
NonBon 3 1 0 0 0.000 0.000 0.500 0.000 0.000 -1.2
"""
    fname = 'GauNonBon.gjf'
    with open(fname, 'w') as fout:
        fout.write(header_gjf.format(f_chg = formal_chg, mult=multi))

        for element, type_, coords, charge in zip(elements, types, coordinates, charges):
            coord_str = '  '.join(f'{x:.6f}' for x in coords)
            fout.write(f'{element}-{type_}-{charge}  {coord_str}\n')

        fout.write(master_func)
        fout.write('! SMARTFIELD FF\n')
        fout.write('!Bonds\n')
        for bond_type, bond_length in zip(bond_types, bond_lengths):
            fout.write(f'HrmStr1 {bond_type}  0.0 {bond_length:.3f}\n')

        fout.write('!Angles\n')
        for angle_type, angle_length in zip(angle_types, angle_lengths):
            fout.write(f'HrmBnd1 {angle_type}  0.0 {angle_length:.3f}\n')
            
        fout.write('!Torsions\n')
        for i, (torsion_type, hybrid) in enumerate(zip(torsion_type_list, hybrid_list)):
            if hybrid == 4.0 or hybrid == 2.0:
               formatted_phase = '0 180 0 0 '
               fout.write(f'AmbTrs {torsion_type} {formatted_phase} 0. 0. 0. 0. {float(hybrid)}\n')
            else:
               formatted_phase = '0 0 0 0 '
               fout.write(f'AmbTrs {torsion_type} {formatted_phase} 0. 0. 0. 0. {float(hybrid)}\n')
            
        fout.write('!VDW\n')
        for vdw_params in vdw:
            s = ' '.join(map(str, vdw_params))
            fout.write(f'{s}\n')

        fout.write('\n')
        

def main():
    parser = rdin.commandline_parser3()
    opts = parser.parse_args()
    json_opts = rdin.read_optfile_3(opts.optfile)
    # opts = parser.parse_args()
    f_qm_log = json_opts['files']['log_qm_file']
    f_qm_fchk = json_opts['files']['fchk_qm_file']
    f_atype = json_opts['files']['atype_file']
    formal_chg = json_opts['charge']
    multi = json_opts['multiplicity']
    text_qm_log = pgau.store_any_file(f_qm_log)
    text_qm_fchk = pgau.store_any_file(f_qm_fchk)
    text_atype = pgau.store_any_file(f_atype)
    
    N_atoms, qm_XYZ = pgau.read_XYZ(text_qm_fchk)
    ele_list, atype_list = pgau.read_NamesTypes(text_atype)             
    # ele_list, type_list = pgau.read_NamesTypes(text_qm_log, N_atoms)
    ric_list, force_1D = pgau.read_RicDim_Grad(text_qm_fchk)
    No_ric = ric_list[0]
    No_bonds = ric_list[1]
    No_angles = ric_list[2]
    No_dihes = ric_list[3]

    # # Reading in Topology in RIC from log file
    chg = pgau.read_CM5(text_qm_log, N_atoms)
    bond_list, angle_list, tors_list = pgau.read_Top(text_qm_log, ric_list)
    k_bonds = np.ones(No_bonds)                          
    k_angles = np.ones(No_angles)
    k_tors = np.ones(No_dihes)
    # print(No_dihes)
    
    mdin = json_opts['opt']
    mode = json_opts['mode']
    bond_type_list, bond_arr, k_bond_arr = fc.set_bonds(qm_XYZ, None, atype_list, \
                      bond_list, k_bonds, mdin, mode)
    angle_type_list, angle_arr, k_angle_arr = fc.set_angles(qm_XYZ, None, atype_list, \
                      angle_list, k_angles, mdin, mode)
    tors_type_list, v1, _, _, _, _, periodic_list = fc.set_torsion(qm_XYZ, atype_list, tors_list, k_tors, force_1D, mode)
    
    # Take out Mirror atom types of bonds & angles
    bond_reduced, _ = aat.make_list_unique(bond_type_list, k_bond_arr)
    angle_reduced, _ = aat.make_list_unique(angle_type_list, k_angle_arr)
    tors_reduced, hybrid_unique = aat.make_list_unique(tors_type_list, periodic_list)
    
    # Print all into Gaussian Input
    print_GauHarm(ele_list, atype_list, qm_XYZ, \
                 bond_reduced, k_bond_arr, bond_arr, \
                 angle_reduced, k_angle_arr, angle_arr, \
                 tors_reduced, hybrid_unique, \
                 chg, formal_chg, multi)
    
    path = os.environ.get("g09root") + "/g09"
    VDW_list = pgau.read_AmberParm(path, atype_list)
    
    print_GauNonBon(ele_list, atype_list, qm_XYZ, \
                 bond_reduced, bond_arr, \
                 angle_reduced, k_angle_arr, \
                 tors_reduced, hybrid_unique, \
                 chg, VDW_list, formal_chg, multi)




if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import parser_gau as pgau
import force_constant_mod as fc
import average_across_types as aat
# import printout_mod as pout
import os
import readin_opts as rdin
import print_top as top
import log2topol
import numpy as np
# import scipy.optimize as optimize
from scipy.sparse import rand
from scipy.optimize import lsq_linear


def get_DiagMatrix(AM):
    for i in range(len(AM)):
        for j in range(len(AM)):
            if i != j :
                 AM[i,j] = 0.0
    return AM


def solve_lsq(A, b):
    rng = np.random.default_rng()
    b = rng.standard_normal(A.shape[0])
    lb = rng.standard_normal(b.shape[0])
    ub = lb + 1
    res = lsq_linear(A, b, bounds=(lb, ub), lsmr_tol='auto', verbose=1)
    print(res)

def fit_hessian(qm_hessian, md_hessian):
    # Extract 3x3 submatrices
    N_atoms = np.int(qm_hessian.shape[0]/3)
    subs_mm, subs_qm = [], []
    for i in range(N_atoms-1):
        for j in range(i+1, N_atoms):
            sub_mm = md_hessian[3*i:3*(i+1), 3*j:3*(j+1)]
            sub_qm = qm_hessian[3*i:3*(i+1), 3*j:3*(j+1)]
            # print(f'{np.array2string(submatrix, precision=2, suppress_small=True)}')
            subs_mm.append(sub_mm)
            subs_qm.append(sub_qm)
            
    # Display extracted submatrices
    for idx, (qm, mm) in enumerate(zip(subs_qm, subs_mm), start=0):
        print(f"Submatrix {idx}:")
        # print(submatrix)
        print(np.array2string(mm, precision=2, suppress_small=True))
        # print(np.linalg.lstsq(mm.flatten(), qm.flatten(), rcond=-1)[0])
        fit = lsq_linear(qm.reshape(-1,1), mm.flatten()).x
        print(fit)


def main():
    parser = rdin.commandline_parser1()
    opts = parser.parse_args()
    json_opts = rdin.read_optfile(opts.optfile)
    f_qm_log = json_opts['files']['log_qm_file']
    f_qm_fchk = json_opts['files']['fchk_qm_file']
    f_atype = json_opts['files']['atype_file']
    f_mm_fchk = json_opts['files']['fchk_mm_file']
    f_nb_fchk = json_opts['files']['fchk_nb_file']
    formal_chg = json_opts['charge']
    multi = json_opts['multiplicity']
    
    # Store all fiels in texts
    text_qm_log = pgau.store_any_file(f_qm_log)
    text_qm_fchk = pgau.store_any_file(f_qm_fchk)
    text_atype = pgau.store_any_file(f_atype)
    text_mm_fchk = pgau.store_any_file(f_mm_fchk)
    text_nb_fchk = pgau.store_any_file(f_nb_fchk)
    
    # Opening texts contents : XYZ, Grad, Hess, Topology & etc
    N_atoms, qm_XYZ = pgau.read_XYZ(text_qm_fchk)                 
    ele_list, atype_list = pgau.read_NamesTypes(text_atype)
    ric_list, force_1D = pgau.read_RicDim_Grad(text_qm_fchk)
    No_ric = ric_list[0]
    No_bonds = ric_list[1]
    No_angles = ric_list[2]
    No_dihes = ric_list[3]

    # Reading in Topology in RIC from log file
    charge = pgau.read_CM5(text_qm_log, N_atoms)
    bond_list, angle_list, tors_list = pgau.read_Top(text_qm_log, ric_list)

    if json_opts['opt'] == 'modsem':
        # Reading XYZ Hessians
        hessXYZ_qm = pgau.read_HessXYZ(text_qm_fchk, N_atoms)
        hessXYZ_nb = pgau.read_HessXYZ(text_nb_fchk, N_atoms)
        # hessXYZ_mm = pgau.read_HessXYZ(text_mm_fchk, N_atoms)
        hess_eff = hessXYZ_qm - hessXYZ_nb 
        k_bonds = np.empty(len(bond_list))
        k_angles = np.empty(len(angle_list))
        k_tors = np.empty(len(tors_list))
        diag_QM = np.diagonal(hessXYZ_qm)  
    else:
        # Reading RIC Hessians 
        hessRIC_qm = pgau.read_HessRIC(text_qm_fchk, ric_list)
        hessRIC_mm = pgau.read_HessRIC(text_mm_fchk, ric_list)
        hessRIC_nb = pgau.read_HessRIC(text_nb_fchk, ric_list)
        hess_eff = hessRIC_qm - hessRIC_nb
        diag_QM = np.diagonal(hess_eff)                       # Take diagonal items of H_QM
        MM_diag = get_DiagMatrix(hessRIC_mm)                  # Make sure H_MM is diagonal
        coeffs = np.linalg.solve(MM_diag, diag_QM)            # Solve Linear System for Bond and Angles only H_MM*K = H_QM ; ignoring Torsion 
        k_bonds = coeffs[0 : No_bonds]                        #  * ((627.509391)/(0.529117*0.529117))                       
        k_angles = coeffs[No_bonds : No_bonds + No_angles ]   #* (627.509391)
        k_tors = coeffs[No_ric - No_dihes : No_ric]           #* 627.509391  # Torsional Gradient; kcal/mol rad
    
    # If opt == sem
    mdin = json_opts['opt']
    mode = json_opts['mode']
    bond_type_list, bond_arr, k_bond_arr = fc.set_bonds(qm_XYZ, hess_eff, atype_list, bond_list, k_bonds, mdin, mode)
    angle_type_list, angle_arr, k_angle_arr = fc.set_angles(qm_XYZ, hess_eff, atype_list, angle_list, k_angles, mdin, mode)
    tors_type_list, v1, v2, v3, tors_arr, phase, periodic_list = fc.set_torsion(qm_XYZ, atype_list, tors_list, k_tors, force_1D, mode)
    
    # Take out mirrored atom types of bonds & angles
    bonds_unique, k_bonds_unique = aat.make_list_unique(bond_type_list, k_bond_arr)
    angles_unique, k_angles_unique = aat.make_list_unique(angle_type_list, k_angle_arr)
    # Same for torsion...
    tors_unique, val_unique = aat.make_list_unique(tors_type_list, v1)

    # Print all into Gaussian Input
    top.print_GauInp(ele_list, atype_list, qm_XYZ, \
                 bonds_unique, k_bonds_unique, bond_arr, \
                 angles_unique, k_angles_unique, angle_arr, \
                 tors_unique, v1, v2, v3, phase, periodic_list, charge, \
                 formal_chg, multi)

    top.print_AmbFrcmod(atype_list, \
                 bonds_unique, k_bonds_unique, bond_arr, \
                 angles_unique, k_angles_unique, angle_arr, \
                 tors_type_list, v1, v2, v3, phase, periodic_list)
    
   # Make dihedral directory for subsquent torsion fitting
    fname = 'SmartField4gau.gjf'
    top.build_dihe_folder(fname, atype_list, charge)
    
    # Build topolo.txt file 
    bond_list, angle_list, tors_list = log2topol.read_Top(text_qm_log)
    os.chdir('dihedrals')
    log2topol.print_topol(bond_list, angle_list, tors_list)
    
    

if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import json
import os
import argparse

def commandline_parser1():
    parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('optfile', nargs='?', help='option file in json')
    # parser.add_argument('-m', '--mode', choices=['modsem', 'ric'],
                        # default='ric', help='method to compute harmonic factors')
    return parser

def dir_path(string):
    if os.path.isdir(string):
        return string
    else:
        raise NotADirectoryError(string)

def commandline_parser2():
    parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter)
    # requiredNamed = parser.add_argument_group('mandatory arguments')
    parser.add_argument('optfile', nargs='?', help='option file in json')
    # requiredNamed.add_argument('-f1','--log_file', help='Gaussian QM log file ')
    # requiredNamed.add_argument('-f2','--fchk_file', help='Gaussain QM fchk file')
    parser.add_argument('-m', '--mode', choices=['all', 'mean'],
                        default='mean', help='averaging across same types; default = mean')
    txt = """path/to/amber.prm in Gaussain root directory
default = current directory """
    parser.add_argument('-path', nargs='?', help=txt, default=os.getcwd(), type=dir_path)
    return parser


def commandline_parser3():
    parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('optfile', nargs='?', help='option file in json')
    parser.add_argument('--path', default='$g09root', help='path to Gaussian09/16 program')
    return parser

def read_optfile(fname):
    """
    Read the option file. expected in json style
    """
    if not os.path.exists(str(fname)):
        raise FileNotFoundError('Missing JSON file')
    with open(fname, 'r')  as fopen:
        data = json.load(fopen)
    for i in ["log_qm_file", "fchk_qm_file", "atype_file", "fchk_mm_file", "fchk_nb_file"]:
        if not os.path.exists(data['files'][i]):
            raise FileNotFoundError(f'Missing {i} file')
    return data

def read_optfile_2(fname):
    """
    Read the option file. expected in json style
    """
    if not os.path.exists(str(fname)):
        raise FileNotFoundError('Missing JSON file')
    with open(fname, 'r')  as fopen:
        data = json.load(fopen)
    for i in ["atom2type", "force_file", "file_xyz", "topol"]:
        if not os.path.exists(data['files'][i]):
            raise FileNotFoundError(f'Missing {i} file')
    return data

def read_optfile_3(fname):
    """
    Read the option file. expected in json style
    """
    if not os.path.exists(str(fname)):
        raise FileNotFoundError('Missing JSON file')
    with open(fname, 'r')  as fopen:
        data = json.load(fopen)
    for i in ["log_qm_file", "fchk_qm_file"]:
        if not os.path.exists(data['files'][i]):
            raise FileNotFoundError(f'Missing {i} file')
    return data


#!/usr/bin/env python3

import numpy as np
np.seterr(divide='ignore', invalid='ignore')

def get_one_hot(values, indexes):
    one_hot = np.eye(np.max(indexes) + 1)[indexes]
    counts = np.sum(one_hot, axis=0)
    average = np.sum((one_hot.T * values), axis=1) / counts
    average = average[~np.isnan(average)]             # Remove NaN 
    return average

def reduce_bond_list(bond_list):
    splitted = list()
    [ splitted.append(i.split()) for i in bond_list]
    data = { (tuple(sorted(item))) for item in splitted}
    new = list(map(list, data))
    reduced = list(' '.join(item) for item in new)
    return reduced

def make_list_unique(var_list, k_values):
    splitted = [ i.split() for i in var_list]
    indexes = [i for i in range(len(splitted))]
    # Get indexes to be removed
    id2del = list()
    for i in range(len(splitted[:])-1):
        for j in range(i+1,len(splitted[:])):
            if (splitted[i] == splitted[j][::-1]) and (i !=j) :
                id2del.append(j)
                indexes[j] = indexes[i]
                break

    # pop indexes out; change res_bonds name 
    if len(id2del) != 0:
        j = id2del[0]
        res_bonds = splitted.copy()
        for n in range(len(id2del)):
            j = id2del[n] - n
            res_bonds.pop(j)
        reduced = list(' '.join(item) for item in res_bonds)
    
        indexes = np.array(indexes)
        k_values_ave = get_one_hot(k_values, indexes)
        return reduced, k_values_ave
    else:
        # print("No duplicates have been found")
        return var_list, k_values






#!/usr/bin/env python3

import sys

def change_mm_COM(filename, angle):
    with open(filename, 'r') as f:
        text = []
        for line in f:
            text.append(line.split())

    # Break it in parts
    first_bit = []
    second_bit = []
    third_bit = []
    for i in text:
        first_bit.append(i)
        if len(i) == 2:
            break
    for i in text[8:]:
        second_bit.append(i)
        if i[0] == 'Variables:':
            break

    second_bit[angle[-1]-1][1] = angle[2]  # Replace Bond
    second_bit[angle[-1]-1][3] = angle[1]  #  .... Angle
    second_bit[angle[-1]-1][5] = angle[0]  #  .. Dihedral

    r1 = second_bit[angle[-1]-1][2]        # Bond Paramters
    r2 = second_bit[angle[-1]-1][4]        # Angle ....
    r3 = second_bit[angle[-1]-1][6]        #  Dihedral ...

    for i in text[len(second_bit)+8:]:
        third_bit.append(i)

    third_bit = [x for x in third_bit if x]   # Take out empty spaces in list

    for i, x in enumerate(third_bit):
        if x[0] == r3:
            third_bit[i][1] = '0.000 S 15 24.0 '


    fout = filename[:-4] + '_mod.gjf'
    text_new = first_bit + second_bit + third_bit
    with open(fout, 'w') as f2:
        for i in text_new:
            print(*i, file=f2)
            # print(i, file = fout)


def main():
    filename = sys.argv[1]
    # angle = [2, 1, 5, 8]
    angle = [1, 5, 8, 11]
    change_mm_COM(filename, angle)

    

if __name__ == '__main__':
    main()
        



#!/usr/bin/env python3

import numpy as np


def unit_vector_N(u_BC, u_AB):
    # Calculates unit normal vector which is perpendicular to plane ABC
    cross_product = np.cross(u_BC, u_AB)
    norm_u_N = np.linalg.norm(cross_product)
    u_N = cross_product / norm_u_N
    return u_N

def angle_force_constant(r_AB, r_BC, u_AB, u_BC, \
                         WR_AB, VR_AB, WR_BC, VR_BC):

    u_N = unit_vector_N(u_BC, u_AB)
    u_PA = np.cross(u_N, u_AB)
    u_PA = u_PA/np.linalg.norm(u_PA)
    u_PC = np.cross(u_N, u_BC)
    u_PC = u_PC / np.linalg.norm(u_PC)

    k_PA = 0.0 
    k_PC = 0.0 
    for i in range(0,3):
        # dot_product = abs(dot_product(u_PA, VR_AB[:,i]))
        k_PA = k_PA +  WR_AB[i] * abs(np.dot(u_PA, VR_AB[:,i]))    
        k_PC = k_PC +  WR_BC[i] * abs(np.dot(u_PC, VR_BC[:,i]))    

    k_theta = ( 1 / ( (r_AB**2) * k_PA) ) + ( 1 / ( (r_BC**2) * k_PC) ) 
    k_theta = 1/k_theta
    k_theta = - k_theta #Change to OPLS form
    k_theta = abs(k_theta * 0.5) #Change to OPLS form

    return k_theta

def bond_force_constant(diff_AB, r_AB, WR, VR,):

    unit_vector = diff_AB / r_AB
    k_AB = 0 
    for i in range(0,3):
        k_AB = k_AB +  WR[i]* abs(np.dot(unit_vector, VR[:,i]))
        # k_AB = k_AB + ( WR[i] * dot_product )

    k_AB = - k_AB * 0.5 # Convert to OPLS form

    return k_AB



def get_Sem_FcBonds(i, j, diff_AB, r_AB, hess):
    sub_hess = hess[(i * 3):((i + 1)*3),(j * 3):((j + 1)*3)]
    WR, VR = np.linalg.eig(sub_hess)
    k_AB = bond_force_constant(diff_AB, r_AB, WR, VR)
    sub_hess = hess[(j * 3):((j + 1)*3),(i * 3):((i + 1)*3)]
    WR, VR = np.linalg.eig(sub_hess)
    k_BA = bond_force_constant(diff_AB, r_AB, WR, VR)
    kk = (k_AB + k_BA)/2.0

    return kk

def get_Sem_FcAngles(i, j, k, r_AB, r_BC, u_AB, u_BC, hess):
    sub_hess_AB = hess[(i * 3):((i + 1)*3),(j * 3):((j + 1)*3)]
    WR_AB, VR_AB = np.linalg.eig(sub_hess_AB)
    sub_hess_BC = hess[(j * 3):((j + 1)*3),(k * 3):((k + 1)*3)]
    WR_BC, VR_BC = np.linalg.eig(sub_hess_BC)
    k_angle = angle_force_constant(r_AB, r_BC, u_AB, u_BC, \
                                   WR_AB, VR_AB, WR_BC, VR_BC)
    
    return k_angle

    
    


    



#from .main import main
#
## Call main function when package is imported
#main()

#__all__ = ['main']
#

from .main import main

#from hessfit.main import main

from os.path import dirname, basename, isfile, join
import glob

modules = glob.glob(join(dirname(__file__), "*.py"))
__all__ = [ basename(f)[:-3] for f in modules if isfile(f) and not f.endswith('__init__.py')]



from setuptools import setup, find_packages

VERSION = '0.0.6'
DESCRIPTION = 'A package that allows to derive force field.'

setup(
      version=VERSION,
      packages = find_packages(),
      package_data={'hessfit': ['json_files/*']},
      install_requires=["numpy", "scipy", "pandas"],
      entry_points = {
        'console_scripts': ['hessfit=main:main',]
      },
     )

